@@ -1,4 +1,4 @@
-# Copyright (C) 2011-2016 2ndQuadrant Italia Srl
+# Copyright (C) 2011-2017 2ndQuadrant Limited
 #
 # This file is part of Barman.
 #
@@ -23,20 +23,20 @@ import datetime
 import logging
 import os
 import shutil
-import time
 from glob import glob
 
 import dateutil.parser
 import dateutil.tz
 
 from barman import output, xlog
-from barman.backup_executor import RsyncBackupExecutor, SshCommandException
-from barman.command_wrappers import DataTransferFailure
-from barman.compression import CompressionIncompatibility, CompressionManager
+from barman.backup_executor import PostgresBackupExecutor, RsyncBackupExecutor
+from barman.compression import CompressionManager
 from barman.config import BackupOptions
-from barman.hooks import (AbortedRetryHookScript, HookScriptRunner,
-                          RetryHookScriptRunner)
-from barman.infofile import BackupInfo, UnknownBackupIdException, WalFileInfo
+from barman.exceptions import (AbortedRetryHookScript,
+                               CompressionIncompatibility, SshCommandException,
+                               UnknownBackupIdException)
+from barman.hooks import HookScriptRunner, RetryHookScriptRunner
+from barman.infofile import BackupInfo, WalFileInfo
 from barman.recovery_executor import RecoveryExecutor
 from barman.remote_status import RemoteStatusMixin
 from barman.utils import fsync_dir, human_readable_timedelta, pretty_size
@@ -44,21 +44,6 @@ from barman.utils import fsync_dir, human_readable_timedelta, pretty_size
 _logger = logging.getLogger(__name__)
 
 
-class DuplicateWalFile(Exception):
-    """
-    A duplicate WAL file has been found
-    """
-    pass
-
-
-class MatchingDuplicateWalFile(DuplicateWalFile):
-    """
-    A duplicate WAL file has been found, but it's identical to the one we
-    already have.
-    """
-    pass
-
-
 class BackupManager(RemoteStatusMixin):
     """Manager of the backup archive for a server"""
 
@@ -69,18 +54,29 @@ class BackupManager(RemoteStatusMixin):
         Constructor
         """
         super(BackupManager, self).__init__()
-        self.name = "default"
         self.server = server
         self.config = server.config
         self._backup_cache = None
         self.compression_manager = CompressionManager(self.config, server.path)
         self.executor = None
         try:
-            self.executor = RsyncBackupExecutor(self)
+            if self.config.backup_method == "postgres":
+                self.executor = PostgresBackupExecutor(self)
+            else:
+                self.executor = RsyncBackupExecutor(self)
         except SshCommandException as e:
             self.config.disabled = True
             self.config.msg_list.append(str(e).strip())
 
+    @property
+    def mode(self):
+        """
+        Property defining the BackupInfo mode content
+        """
+        if self.executor:
+            return self.executor.mode
+        return None
+
     def get_available_backups(self, status_filter=DEFAULT_STATUS_FILTER):
         """
         Get a list of available backups
@@ -262,7 +258,8 @@ class BackupManager(RemoteStatusMixin):
                            len(available_backups),
                            minimum_redundancy)
             return
-
+        # Keep track of when the delete operation started.
+        delete_start_time = datetime.datetime.now()
         output.info("Deleting backup %s for server %s",
                     backup.backup_id, self.config.name)
         previous_backup = self.get_previous_backup(backup.backup_id)
@@ -287,8 +284,26 @@ class BackupManager(RemoteStatusMixin):
                 remove_until = next_backup
             elif BackupOptions.CONCURRENT_BACKUP in self.config.backup_options:
                 remove_until = backup
+
+            timelines_to_protect = set()
+            # If remove_until is not set there are no backup left
+            if remove_until:
+                # Retrieve the list of extra timelines that contains at least
+                # a backup. On such timelines we don't want to delete any WAL
+                for value in self.get_available_backups(
+                        BackupInfo.STATUS_ARCHIVING).values():
+                    # Ignore the backup that is being deleted
+                    if value == backup:
+                        continue
+                    timelines_to_protect.add(value.timeline)
+                # Remove the timeline of `remove_until` from the list.
+                # We have enough information to safely delete unused WAL files
+                # on it.
+                timelines_to_protect -= set([remove_until.timeline])
+
             output.info("Delete associated WAL segments:")
-            for name in self.remove_wal_before_backup(remove_until):
+            for name in self.remove_wal_before_backup(remove_until,
+                                                      timelines_to_protect):
                 output.info("\t%s", name)
         # As last action, remove the backup directory,
         # ending the delete operation
@@ -301,44 +316,13 @@ class BackupManager(RemoteStatusMixin):
                          backup.get_basebackup_directory())
             return
         self.backup_cache_remove(backup)
-        output.info("Done")
-
-    def retry_backup_copy(self, target_function, *args, **kwargs):
-        """
-        Execute the target backup copy function, retrying the configured
-        number of times
-
-        :param target_function: the base backup target function
-        :param args: args for the target function
-        :param kwargs: kwargs of the target function
-        :return: the result of the target function
-        """
-        attempts = 0
-        while True:
-            try:
-                # if is not the first attempt, output the retry number
-                if attempts >= 1:
-                    output.warning("Copy of base backup: retry #%s", attempts)
-                # execute the target function for backup copy
-                return target_function(*args, **kwargs)
-            # catch rsync errors
-            except DataTransferFailure as e:
-                # exit condition: if retry number is lower than configured
-                # retry limit, try again; otherwise exit.
-                if attempts < self.config.basebackup_retry_times:
-                    # Log the exception, for debugging purpose
-                    _logger.exception("Failure in base backup copy: %s", e)
-                    output.warning(
-                        "Copy of base backup failed, waiting for next "
-                        "attempt in %s seconds",
-                        self.config.basebackup_retry_sleep)
-                    # sleep for configured time. then try again
-                    time.sleep(self.config.basebackup_retry_sleep)
-                    attempts += 1
-                else:
-                    # if the max number of attempts is reached and
-                    # there is still an error, exit re-raising the exception.
-                    raise
+        # Save the time of the complete removal of the backup
+        delete_end_time = datetime.datetime.now()
+        output.info("Deleted backup %s (start time: %s, elapsed time: %s)",
+                    backup.backup_id,
+                    delete_start_time.ctime(),
+                    human_readable_timedelta(
+                        delete_end_time - delete_start_time))
 
     def backup(self):
         """
@@ -355,7 +339,8 @@ class BackupManager(RemoteStatusMixin):
             backup_info.save()
             self.backup_cache_add(backup_info)
             output.info(
-                "Starting backup for server %s in %s",
+                "Starting backup using %s method for server %s in %s",
+                self.mode,
                 self.config.name,
                 backup_info.get_basebackup_directory())
 
@@ -404,7 +389,11 @@ class BackupManager(RemoteStatusMixin):
                         backup_info.end_xlog,
                         backup_info.end_wal,
                         backup_info.end_offset)
-            output.info("Backup completed")
+            output.info("Backup completed (start time: %s, elapsed time: %s)",
+                        self.executor.copy_start_time,
+                        human_readable_timedelta(
+                            self.executor.copy_end_time
+                            - self.executor.copy_start_time))
             # Create a restore point after a backup
             target_name = 'barman_%s' % backup_info.backup_id
             self.server.postgres.create_restore_point(target_name)
@@ -412,6 +401,10 @@ class BackupManager(RemoteStatusMixin):
             if backup_info:
                 backup_info.save()
 
+                # Make sure we are not holding any PostgreSQL connection
+                # during the post-backup scripts
+                self.server.close()
+
                 # Run the post-backup-retry-script if present.
                 try:
                     retry_script = RetryHookScriptRunner(
@@ -474,9 +467,8 @@ class BackupManager(RemoteStatusMixin):
 
         :param bool verbose: report even if no actions
         """
-        with self.server.xlogdb('a') as fxlogdb:
-            for archiver in self.server.archivers:
-                archiver.archive(fxlogdb, verbose)
+        for archiver in self.server.archivers:
+            archiver.archive(verbose)
 
     def cron_retention_policy(self):
         """
@@ -544,9 +536,10 @@ class BackupManager(RemoteStatusMixin):
                 # We always try to remove the the trailing directories,
                 # this means that hashdir is not empty.
                 pass
-        except OSError:
-            _logger.warning('Expected WAL file %s not found during delete',
-                            wal_info.name, exc_info=1)
+        except OSError as e:
+            output.warning('Ignoring deletion of WAL file %s '
+                           'for server %s: %s',
+                           wal_info.name, self.config.name, e)
 
     def check(self, check_strategy):
         """
@@ -555,10 +548,10 @@ class BackupManager(RemoteStatusMixin):
         :param CheckStrategy check_strategy: the strategy for the management
              of the results of the various checks
         """
+        check_strategy.init_check('compression settings')
         # Check compression_setting parameter
         if self.config.compression and not self.compression_manager.check():
-            check_strategy.result(self.config.name,
-                                  'compression settings', False)
+            check_strategy.result(self.config.name, False)
         else:
             status = True
             try:
@@ -567,19 +560,18 @@ class BackupManager(RemoteStatusMixin):
                 check_strategy.result(self.config.name,
                                       '%s setting' % field, False)
                 status = False
-            check_strategy.result(self.config.name,
-                                  'compression settings', status)
+            check_strategy.result(self.config.name, status)
 
         # Failed backups check
+        check_strategy.init_check('failed backups')
         failed_backups = self.get_available_backups((BackupInfo.FAILED,))
         status = len(failed_backups) == 0
         check_strategy.result(
             self.config.name,
-            'failed backups',
             status,
-            'there are %s failed backups' % (len(failed_backups,))
+            hint='there are %s failed backups' % (len(failed_backups,))
         )
-
+        check_strategy.init_check('minimum redundancy requirements')
         # Minimum redundancy checks
         no_backups = len(self.get_available_backups())
         # Check minimum_redundancy_requirements parameter
@@ -588,9 +580,8 @@ class BackupManager(RemoteStatusMixin):
         else:
             status = True
         check_strategy.result(
-            self.config.name,
-            'minimum redundancy requirements', status,
-            'have %s backups, expected at least %s' % (
+            self.config.name, status,
+            hint='have %s backups, expected at least %s' % (
                 no_backups, self.config.minimum_redundancy))
 
         # TODO: Add a check for the existence of ssh and of rsync
@@ -685,6 +676,12 @@ class BackupManager(RemoteStatusMixin):
                                     wal_count += 1
                                 elif xlog.is_backup_file(fullname):
                                     label_count += 1
+                                elif fullname.endswith('.tmp'):
+                                    _logger.warning(
+                                        'temporary file found '
+                                        'rebuilding the wal database: %s',
+                                        fullname)
+                                    continue
                                 else:
                                     _logger.warning(
                                         'unexpected file '
@@ -715,14 +712,66 @@ class BackupManager(RemoteStatusMixin):
                     '(history: %s, backup_labels: %s, wal_file: %s)',
                     self.config.name, history_count, label_count, wal_count)
 
-    def remove_wal_before_backup(self, backup_info):
+    def get_latest_archived_wals_info(self):
+        """
+        Return a dictionary of timelines associated with the
+        WalFileInfo of the last WAL file in the archive,
+        or None if the archive doesn't contain any WAL file.
+
+        :rtype: dict[str, WalFileInfo]|None
+        """
+        from os.path import isdir, join
+
+        root = self.config.wals_directory
+
+        # If the WAL archive directory doesn't exists the archive is empty
+        if not isdir(root):
+            return None
+
+        # Traverse all the directory in the archive in reverse order,
+        # returning the first WAL file found
+        timelines = {}
+        for name in sorted(os.listdir(root), reverse=True):
+            fullname = join(root, name)
+            # All relevant files are in subdirectories, so
+            # we skip any non-directory entry
+            if isdir(fullname):
+                # Extract the timeline. If it is not valid, skip this directory
+                try:
+                    timeline = name[0:8]
+                    int(timeline, 16)
+                except ValueError:
+                    continue
+
+                # If this timeline already has a file, skip this directory
+                if timeline in timelines:
+                    continue
+
+                hash_dir = fullname
+                # Inspect contained files in reverse order
+                for wal_name in sorted(os.listdir(hash_dir), reverse=True):
+                    fullname = join(hash_dir, wal_name)
+                    # Return the first file that has the correct name
+                    if not isdir(fullname) and xlog.is_wal_file(fullname):
+                        timelines[timeline] = WalFileInfo.from_file(fullname)
+                        break
+
+        # Return the timeline map or None if it is empty
+        return timelines or None
+
+    def remove_wal_before_backup(self, backup_info, timelines_to_protect=None):
         """
         Remove WAL files which have been archived before the start of
         the provided backup.
 
         If no backup_info is provided delete all available WAL files
 
+        If timelines_to_protect list is passed, never remove a wal in one of
+        these timelines.
+
         :param BackupInfo|None backup_info: the backup information structure
+        :param set timelines_to_protect: optional list of timelines
+            to protect
         :return list: a list of removed WAL files
         """
         removed = []
@@ -738,13 +787,25 @@ class BackupManager(RemoteStatusMixin):
                             "to solve this issue",
                             wal_info.name, self.config.name)
                         continue
-                    # Keeps the WAL segment if it is a history file or later
+
+                    # Keeps the WAL segment if it is a history file
+                    keep = xlog.is_history_file(wal_info.name)
+
+                    # Keeps the WAL segment if its timeline is in
+                    # `timelines_to_protect`
+                    if timelines_to_protect:
+                        tli, _, _ = xlog.decode_segment_name(wal_info.name)
+                        keep |= tli in timelines_to_protect
+
+                    # Keeps the WAL segment if it is a newer
                     # than the given backup (the first available)
-                    if (xlog.is_history_file(wal_info.name) or
-                            (backup_info and
-                                wal_info.name >= backup_info.begin_wal)):
+                    if backup_info:
+                        keep |= wal_info.name >= backup_info.begin_wal
+
+                    # If the file has to be kept write it in the new xlogdb
+                    # otherwise delete it  and record it in the removed list
+                    if keep:
                         fxlogdb_new.write(wal_info.to_xlogdb_line())
-                        continue
                     else:
                         self.delete_wal(wal_info)
                         removed.append(wal_info.name)
