@@ -1,4 +1,4 @@
-# Copyright (C) 2011-2016 2ndQuadrant Italia Srl
+# Copyright (C) 2011-2017 2ndQuadrant Limited
 #
 # This file is part of Barman.
 #
@@ -23,20 +23,41 @@ import logging
 import os
 import sys
 from argparse import SUPPRESS, ArgumentTypeError
+from contextlib import closing
 
 from argh import ArghParser, arg, expects_obj, named
 
 import barman.config
 import barman.diagnose
 from barman import output
+from barman.config import RecoveryOptions
+from barman.exceptions import BadXlogSegmentName
 from barman.infofile import BackupInfo
 from barman.server import Server
 from barman.utils import configure_logging, drop_privileges, parse_log_level
-from barman.xlog import BadXlogSegmentName
 
 _logger = logging.getLogger(__name__)
 
 
+def check_non_negative(value):
+    """
+    Check for a positive integer option
+
+    :param value: str containing the value to check
+    """
+    if value is None:
+        return None
+    try:
+        int_value = int(value)
+    except Exception:
+        raise ArgumentTypeError("'%s' is not a valid non negative integer" %
+                                value)
+    if int_value < 0:
+        raise ArgumentTypeError("'%s' is not a valid non negative integer" %
+                                value)
+    return int_value
+
+
 def check_positive(value):
     """
     Check for a positive integer option
@@ -48,9 +69,11 @@ def check_positive(value):
     try:
         int_value = int(value)
     except Exception:
-        raise ArgumentTypeError("'%s' is not a valid positive integer" % value)
-    if int_value < 0:
-        raise ArgumentTypeError("'%s' is not a valid positive integer" % value)
+        raise ArgumentTypeError("'%s' is not a valid positive integer" %
+                                value)
+    if int_value < 1:
+        raise ArgumentTypeError("'%s' is not a valid positive integer" %
+                                value)
     return int_value
 
 
@@ -155,12 +178,15 @@ def backup_completer(prefix, parsed_args, **kwargs):
           'If no argument is given "link" is assumed')
 @arg('--retry-times',
      help='Number of retries after an error if base backup copy fails.',
-     type=check_positive)
+     type=check_non_negative)
 @arg('--retry-sleep',
      help='Wait time after a failed base backup copy, before retrying.',
-     type=check_positive)
+     type=check_non_negative)
 @arg('--no-retry', help='Disable base backup copy retry logic.',
      dest='retry_times', action='store_const', const=0)
+@arg('--jobs', '-j',
+     help='Run the copy in parallel using NJOBS processes.',
+     type=check_positive, metavar='NJOBS')
 @expects_obj
 def backup(args):
     """
@@ -182,7 +208,10 @@ def backup(args):
             server.config.basebackup_retry_times = args.retry_times
         if hasattr(args, 'immediate_checkpoint'):
             server.config.immediate_checkpoint = args.immediate_checkpoint
-        server.backup()
+        if args.jobs is not None:
+            server.config.parallel_jobs = args.jobs
+        with closing(server):
+            server.backup()
     output.close_and_exit()
 
 
@@ -206,7 +235,8 @@ def list_backup(args):
             continue
 
         output.init('list_backup', name, minimal=args.minimal)
-        server.list_backups()
+        with closing(server):
+            server.list_backups()
     output.close_and_exit()
 
 
@@ -227,7 +257,40 @@ def status(args):
             continue
 
         output.init('status', name)
-        server.status()
+        with closing(server):
+            server.status()
+    output.close_and_exit()
+
+
+@named('replication-status')
+@arg('server_name', nargs='+',
+     completer=server_completer_all,
+     help='specifies the server name for the command')
+@arg('--minimal', help='machine readable output', action='store_true')
+@arg('--target', choices=('all', 'hot-standby', 'wal-streamer'),
+     default='all',
+     help='''
+         Possible values are: 'hot-standby' (only hot standby servers),
+         'wal-streamer' (only WAL streaming clients, such as pg_receivexlog),
+         'all' (any of them). Defaults to %(default)s''')
+@expects_obj
+def replication_status(args):
+    """
+    Shows live information and status of any streaming client
+    """
+    servers = get_server_list(args, skip_inactive=True)
+    for name in sorted(servers):
+        server = servers[name]
+
+        # Skip the server (apply general rule)
+        if not manage_server_command(server, name):
+            continue
+
+        with closing(server):
+            output.init('replication_status',
+                        name,
+                        minimal=args.minimal)
+            server.replication_status(args.target)
     output.close_and_exit()
 
 
@@ -247,14 +310,15 @@ def rebuild_xlogdb(args):
         if not manage_server_command(server, name):
             continue
 
-        server.rebuild_xlogdb()
+        with closing(server):
+            server.rebuild_xlogdb()
     output.close_and_exit()
 
 
 @arg('server_name',
      completer=server_completer,
      help='specifies the server name for the command')
-@arg('--target-tli', help='target timeline', type=int)
+@arg('--target-tli', help='target timeline', type=check_positive)
 @arg('--target-time',
      help='target time. You can use any valid unambiguous representation. '
           'e.g: "YYYY-MM-DD HH:MM:SS.mmm"')
@@ -281,12 +345,25 @@ def rebuild_xlogdb(args):
      help='the directory where the new server is created')
 @arg('--retry-times',
      help='Number of retries after an error if base backup copy fails.',
-     type=check_positive)
+     type=check_non_negative)
 @arg('--retry-sleep',
      help='Wait time after a failed base backup copy, before retrying.',
-     type=check_positive)
+     type=check_non_negative)
 @arg('--no-retry', help='Disable base backup copy retry logic.',
      dest='retry_times', action='store_const', const=0)
+@arg('--jobs', '-j',
+     help='Run the copy in parallel using NJOBS processes.',
+     type=check_positive, metavar='NJOBS')
+@arg('--get-wal',
+     help='Enable the get-wal option during the recovery.',
+     dest='get_wal',
+     action='store_true',
+     default=SUPPRESS)
+@arg('--no-get-wal',
+     help='Disable the get-wal option during recovery.',
+     dest='get_wal',
+     action='store_false',
+     default=SUPPRESS)
 @expects_obj
 def recover(args):
     """
@@ -341,15 +418,23 @@ def recover(args):
         server.config.basebackup_retry_sleep = args.retry_sleep
     if args.retry_times is not None:
         server.config.basebackup_retry_times = args.retry_times
-    server.recover(backup_id,
-                   args.destination_directory,
-                   tablespaces=tablespaces,
-                   target_tli=args.target_tli,
-                   target_time=args.target_time,
-                   target_xid=args.target_xid,
-                   target_name=args.target_name,
-                   exclusive=args.exclusive,
-                   remote_command=args.remote_ssh_command)
+    if hasattr(args, 'get_wal'):
+        if args.get_wal:
+            server.config.recovery_options.add(RecoveryOptions.GET_WAL)
+        else:
+            server.config.recovery_options.remove(RecoveryOptions.GET_WAL)
+    if args.jobs is not None:
+        server.config.parallel_jobs = args.jobs
+    with closing(server):
+        server.recover(backup_id,
+                       args.destination_directory,
+                       tablespaces=tablespaces,
+                       target_tli=args.target_tli,
+                       target_time=args.target_time,
+                       target_xid=args.target_xid,
+                       target_name=args.target_name,
+                       exclusive=args.exclusive,
+                       remote_command=args.remote_ssh_command)
 
     output.close_and_exit()
 
@@ -364,16 +449,61 @@ def show_server(args):
     """
     Show all configuration parameters for the specified servers
     """
-    servers = get_server_list(args, skip_inactive=True)
+    servers = get_server_list(args)
     for name in sorted(servers):
         server = servers[name]
 
         # Skip the server (apply general rule)
-        if not manage_server_command(server, name):
+        if not manage_server_command(
+                server, name, skip_inactive=False,
+                skip_disabled=False, disabled_is_error=False):
             continue
 
+        # If the server has been manually disabled
+        if not server.config.active:
+            name += " (inactive)"
+        # If server has configuration errors
+        elif server.config.disabled:
+            name += " (WARNING: disabled)"
         output.init('show_server', name)
-        server.show()
+        with closing(server):
+            server.show()
+    output.close_and_exit()
+
+
+@named('switch-xlog')
+@arg('server_name', nargs='+',
+     completer=server_completer_all,
+     help="specifies the server name target of the switch-xlog command")
+@arg('--force',
+     help='forces the switch of a xlog by executing a checkpoint before',
+     dest='force',
+     action='store_true',
+     default=False)
+@arg('--archive',
+     help='wait for one xlog file to be archived',
+     dest='archive',
+     action='store_true',
+     default=False)
+@arg('--archive-timeout',
+     help='the time, in seconds, the archiver will wait for a new xlog file '
+          'to be archived before timing out',
+     metavar='TIMEOUT',
+     default='30',
+     type=check_non_negative)
+@expects_obj
+def switch_xlog(args):
+    """
+    Execute the switch-xlog command on the target server
+    """
+    servers = get_server_list(args, skip_inactive=True)
+    for name in sorted(servers):
+        server = servers[name]
+        # Skip the server (apply general rule)
+        if not manage_server_command(server, name):
+            continue
+        with closing(server):
+            server.switch_xlog(args.force, args.archive, args.archive_timeout)
     output.close_and_exit()
 
 
@@ -409,7 +539,8 @@ def check(args):
         elif server.config.disabled:
             name += " (WARNING: disabled)"
         output.init('check', name, server.config.active)
-        server.check()
+        with closing(server):
+            server.check()
     output.close_and_exit()
 
 
@@ -441,7 +572,8 @@ def show_backup(args):
 
     # Retrieves the backup
     backup_info = parse_backup_id(server, args)
-    server.show_backup(backup_info)
+    with closing(server):
+        server.show_backup(backup_info)
     output.close_and_exit()
 
 
@@ -468,9 +600,9 @@ def list_files(args):
     server = get_server(args)
 
     # Retrieves the backup
-    backup_id = parse_backup_id(server, args)
+    backup_info = parse_backup_id(server, args)
     try:
-        for line in backup_id.get_list_of_files(args.target):
+        for line in backup_info.get_list_of_files(args.target):
             output.info(line, log=False)
     except BadXlogSegmentName as e:
         output.error(
@@ -496,7 +628,8 @@ def delete(args):
 
     # Retrieves the backup
     backup_id = parse_backup_id(server, args)
-    server.delete_backup(backup_id)
+    with closing(server):
+        server.delete_backup(backup_id)
     output.close_and_exit()
 
 
@@ -510,12 +643,20 @@ def delete(args):
      help='put the retrieved WAL file in this directory '
           'with the original name',
      default=SUPPRESS)
-@arg('--gzip', '-x',
+@arg('--gzip', '-z', '-x',
      help='compress the output with gzip',
      action='store_const', const='gzip', dest='compression', default=SUPPRESS)
 @arg('--bzip2', '-j',
      help='compress the output with bzip2',
      action='store_const', const='bzip2', dest='compression', default=SUPPRESS)
+@arg('--peek', '-p',
+     help="peek from the WAL archive up to 'SIZE' WAL files, starting "
+          "from the requested one. 'SIZE' must be an integer >= 1. "
+          "When invoked with this option, get-wal returns a list of "
+          "zero to 'SIZE' WAL segment names, one per row.",
+     metavar='SIZE',
+     type=check_positive,
+     default=SUPPRESS)
 @expects_obj
 def get_wal(args):
     """
@@ -523,15 +664,20 @@ def get_wal(args):
     The content will be streamed on standard output unless
     the --output-directory option is specified.
     """
-    server = get_server(args)
+    server = get_server(args, inactive_is_error=True)
+
     # Retrieve optional arguments. If an argument is not specified,
     # the namespace doesn't contain it due to SUPPRESS default.
     # In that case we pick 'None' using getattr third argument.
     compression = getattr(args, 'compression', None)
     output_directory = getattr(args, 'output_directory', None)
-    server.get_wal(args.wal_name,
-                   compression=compression,
-                   output_directory=output_directory)
+    peek = getattr(args, 'peek', None)
+
+    with closing(server):
+        server.get_wal(args.wal_name,
+                       compression=compression,
+                       output_directory=output_directory,
+                       peek=peek)
     output.close_and_exit()
 
 
@@ -548,7 +694,8 @@ def archive_wal(args):
 
     """
     server = get_server(args)
-    server.archive_wal()
+    with closing(server):
+        server.archive_wal()
     output.close_and_exit()
 
 
@@ -558,6 +705,10 @@ def archive_wal(args):
 @arg('--reset', help='reset the status of receive-wal removing '
                      'any status files',
      action='store_true')
+@arg('--create-slot', help='create the replication slot, if it does not exist',
+     action='store_true')
+@arg('--drop-slot', help='drop the replication slot, if it exists',
+     action='store_true')
 @arg('server_name',
      completer=server_completer,
      help='specifies the server name for the command')
@@ -575,8 +726,15 @@ def receive_wal(args):
     # termination signal, otherwise attempt to start it
     elif args.stop:
         server.kill('receive-wal')
+    elif args.create_slot:
+        with closing(server):
+            server.create_physical_repslot()
+    elif args.drop_slot:
+        with closing(server):
+            server.drop_repslot()
     else:
-        server.receive_wal(reset=args.reset)
+        with closing(server):
+            server.receive_wal(reset=args.reset)
     output.close_and_exit()
 
 
@@ -648,6 +806,7 @@ def global_config(args):
 
 
 def get_server(args, skip_inactive=True, skip_disabled=False,
+               inactive_is_error=False,
                on_error_stop=True, suppress_error=False):
     """
     Get a single server retrieving its configuration (wraps get_server_list())
@@ -660,8 +819,9 @@ def get_server(args, skip_inactive=True, skip_disabled=False,
     :param args: an argparse namespace containing a single
         server_name parameter
         WARNING: the function modifies the content of this parameter
-    :param bool skip_inactive: skip inactive servers when 'all' is required
-    :param bool skip_disabled: skip disabled servers when 'all' is required
+    :param bool skip_inactive: do nothing if the server is inactive
+    :param bool skip_disabled: do nothing if the server is disabled
+    :param bool inactive_is_error: treat inactive server as error
     :param bool on_error_stop: stop if an error is found
     :param bool suppress_error: suppress display of errors (e.g. diagnose)
     :rtype: barman.server.Server|None
@@ -681,6 +841,10 @@ def get_server(args, skip_inactive=True, skip_disabled=False,
     # Builds a list from a single given name
     args.server_name = [name]
 
+    # Skip_inactive is reset if inactive_is_error is set, because
+    # it needs to retrieve the inactive server to emit the error.
+    skip_inactive &= not inactive_is_error
+
     # Retrieve the requested server
     servers = get_server_list(args, skip_inactive, skip_disabled,
                               on_error_stop, suppress_error)
@@ -698,7 +862,9 @@ def get_server(args, skip_inactive=True, skip_disabled=False,
     # Apply standard validation control and skips
     # the server if inactive or disabled, displaying standard
     # error messages. If on_error_stop (default) exits
-    if not manage_server_command(server, name) and on_error_stop:
+    if not manage_server_command(server, name,
+                                 inactive_is_error) and \
+            on_error_stop:
         output.close_and_exit()
         # The following return statement will never be reached
         # but it is here for clarity
@@ -860,9 +1026,10 @@ def main():
     """
     The main method of Barman
     """
-    p = ArghParser()
+    p = ArghParser(epilog='Barman by 2ndQuadrant (www.2ndQuadrant.com)')
     p.add_argument('-v', '--version', action='version',
-                   version=barman.__version__)
+                   version='%s\n\nBarman by 2ndQuadrant (www.2ndQuadrant.com)'
+                           % barman.__version__)
     p.add_argument('-c', '--config',
                    help='uses a configuration file '
                         '(defaults: %s)'
@@ -890,7 +1057,9 @@ def main():
             recover,
             show_backup,
             show_server,
-            status
+            replication_status,
+            status,
+            switch_xlog,
         ]
     )
     # noinspection PyBroadException
@@ -898,7 +1067,7 @@ def main():
         p.dispatch(pre_call=global_config)
     except KeyboardInterrupt:
         msg = "Process interrupted by user (KeyboardInterrupt)"
-        output.exception(msg)
+        output.error(msg)
     except Exception as e:
         msg = "%s\nSee log file for more details." % e
         output.exception(msg)
