@@ -1,4 +1,4 @@
-# Copyright (C) 2013-2016 2ndQuadrant Italia Srl
+# Copyright (C) 2013-2017 2ndQuadrant Limited
 #
 # This file is part of Barman.
 #
@@ -21,12 +21,14 @@ This module control how the output of Barman will be rendered
 
 from __future__ import print_function
 
+import datetime
 import inspect
 import logging
 import sys
 
 from barman.infofile import BackupInfo
-from barman.utils import pretty_size
+from barman.utils import human_readable_timedelta, pretty_size
+from barman.xlog import diff_lsn
 
 __all__ = [
     'error_occurred', 'debug', 'info', 'warning', 'error', 'exception',
@@ -438,6 +440,19 @@ class ConsoleOutputWriter(object):
                           assertion.key,
                           assertion.value)
 
+        if results['missing_files']:
+            # At least one file is missing, warn the user
+            self.info("")
+            self.info("WARNING")
+            self.info("The following configuration files have not been "
+                      "saved during backup, hence they have not been "
+                      "restored.")
+            self.info("You need to manually restore them "
+                      "in order to start the recovered PostgreSQL instance:")
+            self.info("")
+            for file_name in results['missing_files']:
+                self.info("    %s" % file_name)
+
         if results['delete_barman_xlog']:
             self.info("")
             self.info("After the recovery, please remember to remove the "
@@ -597,6 +612,25 @@ class ConsoleOutputWriter(object):
             self.info("    Begin time           : %s",
                       data['begin_time'])
             self.info("    End time             : %s", data['end_time'])
+            # If copy statistics are available print a summary
+            copy_stats = data.get('copy_stats')
+            if copy_stats:
+                copy_time = copy_stats.get('copy_time')
+                if copy_time:
+                    value = human_readable_timedelta(
+                        datetime.timedelta(seconds=copy_time))
+                    # Show analysis time if it is more than a second
+                    analysis_time = copy_stats.get('analysis_time')
+                    if analysis_time is not None and analysis_time >= 1:
+                        value += " + %s startup" % (human_readable_timedelta(
+                            datetime.timedelta(seconds=analysis_time)))
+                    self.info("    Copy time            : %s", value)
+                    size = data['deduplicated_size'] or data['size']
+                    value = "%s/s" % pretty_size(size/copy_time)
+                    number_of_workers = copy_stats.get('number_of_workers', 1)
+                    if number_of_workers > 1:
+                        value += " (%s jobs)" % number_of_workers
+                    self.info("    Estimated throughput : %s", value)
             self.info("    Begin Offset         : %s",
                       data['begin_offset'])
             self.info("    End Offset           : %s",
@@ -621,6 +655,11 @@ class ConsoleOutputWriter(object):
                     '{percent:.2%}'.format(
                         percent=data['wal_until_next_compression_ratio']))
             self.info("    Last available       : %s", data['wal_last'])
+            if data['children_timelines']:
+                timelines = data['children_timelines']
+                self.info(
+                    "    Reachable timelines  : %s",
+                    ", ".join([str(history.tli) for history in timelines]))
             self.info("")
             self.info("  Catalog information:")
             self.info("    Retention Policy     : %s",
@@ -632,6 +671,11 @@ class ConsoleOutputWriter(object):
             self.info("    Next Backup          : %s",
                       data.setdefault('next_backup_id', 'not available') or
                       '- (this is the latest base backup)')
+            if data['children_timelines']:
+                self.info("")
+                self.info(
+                    "WARNING: WAL information is inaccurate due to "
+                    "multiple timelines interacting with this backup")
         else:
             if data['error']:
                 self.info("  Error:            : %s",
@@ -662,6 +706,180 @@ class ConsoleOutputWriter(object):
             description=description, message=message))
         self.info("\t%s: %s", description, message)
 
+    def init_replication_status(self, server_name, minimal=False):
+        """
+        Init the 'standby-status' command
+
+        :param str server_name: the server we are start listing
+        :param str minimal: minimal output
+        """
+        self.minimal = minimal
+
+    def result_replication_status(self, server_name, target, xlog_location,
+                                  standby_info):
+        """
+        Record a result line of a server status command
+
+        and output it as INFO
+
+        :param str server_name: the replication server
+        :param str target: all|hot-standby|wal-streamer
+        :param str xlog_location: server's xlog location
+        :param StatReplication standby_info: status info of a standby
+        """
+
+        if target == 'hot-standby':
+            title = 'hot standby servers'
+        elif target == 'wal-streamer':
+            title = 'WAL streamers'
+        else:
+            title = 'streaming clients'
+
+        if self.minimal:
+            # Minimal output
+            if xlog_location:
+                # xlog location from the master
+                self.info("%s for master '%s' (xlog @ %s):",
+                          title.capitalize(), server_name, xlog_location)
+            else:
+                # We are connected to a standby
+                self.info("%s for slave '%s':",
+                          title.capitalize(), server_name)
+        else:
+            # Full output
+            self.info("Status of %s for server '%s':",
+                      title, server_name)
+            # xlog location from the master
+            if xlog_location:
+                self.info("  Current xlog location on master: %s",
+                          xlog_location)
+
+        if standby_info is not None and not len(standby_info):
+            self.info("  No %s attached", title)
+            return
+
+        # Minimal output
+        if self.minimal:
+            n = 1
+            for standby in standby_info:
+                if not standby.replay_location:
+                    # WAL streamer
+                    self.info("  %s. W) %s@%s S:%s W:%s P:%s AN:%s",
+                              n,
+                              standby.usename,
+                              standby.client_addr or 'socket',
+                              standby.sent_location,
+                              standby.write_location,
+                              standby.sync_priority,
+                              standby.application_name)
+                else:
+                    # Standby
+                    self.info("  %s. %s) %s@%s S:%s F:%s R:%s P:%s AN:%s",
+                              n,
+                              standby.sync_state[0].upper(),
+                              standby.usename,
+                              standby.client_addr or 'socket',
+                              standby.sent_location,
+                              standby.flush_location,
+                              standby.replay_location,
+                              standby.sync_priority,
+                              standby.application_name)
+                n += 1
+        else:
+            n = 1
+            self.info("  Number of %s: %s",
+                      title, len(standby_info))
+            for standby in standby_info:
+                self.info("")
+
+                # Calculate differences in bytes
+                sent_diff = diff_lsn(standby.sent_location,
+                                     standby.current_location)
+                write_diff = diff_lsn(standby.write_location,
+                                      standby.current_location)
+                flush_diff = diff_lsn(standby.flush_location,
+                                      standby.current_location)
+                replay_diff = diff_lsn(standby.replay_location,
+                                       standby.current_location)
+
+                # Determine the sync stage of the client
+                sync_stage = None
+                if not standby.replay_location:
+                    client_type = 'WAL streamer'
+                    max_level = 3
+                else:
+                    client_type = 'standby'
+                    max_level = 5
+                    # Only standby can replay WAL info
+                    if replay_diff == 0:
+                        sync_stage = '5/5 Hot standby (max)'
+                    elif flush_diff == 0:
+                        sync_stage = '4/5 2-safe'  # remote flush
+
+                # If not yet done, set the sync stage
+                if not sync_stage:
+                    if write_diff == 0:
+                        sync_stage = '3/%s Remote write' % max_level
+                    elif sent_diff == 0:
+                        sync_stage = '2/%s WAL Sent (min)' % max_level
+                    else:
+                        sync_stage = '1/%s 1-safe' % max_level
+
+                # Synchronous standby
+                if getattr(standby, 'sync_priority', None) > 0:
+                    self.info("  %s. #%s %s %s",
+                              n,
+                              standby.sync_priority,
+                              standby.sync_state.capitalize(),
+                              client_type)
+                # Asynchronous standby
+                else:
+                    self.info("  %s. %s %s",
+                              n,
+                              standby.sync_state.capitalize(),
+                              client_type)
+                self.info("     Application name: %s",
+                          standby.application_name)
+                self.info("     Sync stage      : %s",
+                          sync_stage)
+                if getattr(standby, 'client_addr', None):
+                    self.info("     Communication   : TCP/IP")
+                    self.info("     IP Address      : %s "
+                              "/ Port: %s / Host: %s",
+                              standby.client_addr,
+                              standby.client_port,
+                              standby.client_hostname or '-')
+                else:
+                    self.info("     Communication   : Unix domain socket")
+                self.info("     User name       : %s", standby.usename)
+                self.info("     Current state   : %s (%s)",
+                          standby.state,
+                          standby.sync_state)
+                if getattr(standby, 'slot_name', None):
+                    self.info("     Replication slot: %s", standby.slot_name)
+                self.info("     WAL sender PID  : %s", standby.pid)
+                self.info("     Started at      : %s", standby.backend_start)
+                if getattr(standby, 'backend_xmin', None):
+                    self.info("     Standby's xmin  : %s",
+                              standby.backend_xmin or '-')
+                if getattr(standby, 'sent_location', None):
+                    self.info("     Sent location   : %s (diff: %s)",
+                              standby.sent_location,
+                              pretty_size(sent_diff))
+                if getattr(standby, 'write_location', None):
+                    self.info("     Write location  : %s (diff: %s)",
+                              standby.write_location,
+                              pretty_size(write_diff))
+                if getattr(standby, 'flush_location', None):
+                    self.info("     Flush location  : %s (diff: %s)",
+                              standby.flush_location,
+                              pretty_size(flush_diff))
+                if getattr(standby, 'replay_location', None):
+                    self.info("     Replay location : %s (diff: %s)",
+                              standby.replay_location,
+                              pretty_size(replay_diff))
+                n += 1
+
     def init_list_server(self, server_name, minimal=False):
         """
         Init the list-server command
