@@ -1,4 +1,4 @@
-# Copyright (C) 2011-2016 2ndQuadrant Italia Srl
+# Copyright (C) 2011-2017 2ndQuadrant Limited
 #
 # This file is part of Barman.
 #
@@ -17,20 +17,21 @@
 
 import collections
 import datetime
+import errno
 import filecmp
 import logging
 import os
 import shutil
 from abc import ABCMeta, abstractmethod
-from distutils.version import LooseVersion as Version
 from glob import glob
 
-from barman import output, utils, xlog
-from barman.backup import DuplicateWalFile, MatchingDuplicateWalFile
-from barman.command_wrappers import (Command, CommandFailedException,
-                                     PgReceiveXlog)
-from barman.hooks import (AbortedRetryHookScript, HookScriptRunner,
-                          RetryHookScriptRunner)
+from distutils.version import LooseVersion as Version
+
+from barman import output, xlog
+from barman.command_wrappers import CommandFailedException, PgReceiveXlog
+from barman.exceptions import (AbortedRetryHookScript, ArchiverFailure,
+                               DuplicateWalFile, MatchingDuplicateWalFile)
+from barman.hooks import HookScriptRunner, RetryHookScriptRunner
 from barman.infofile import WalFileInfo
 from barman.remote_status import RemoteStatusMixin
 from barman.utils import fsync_dir, mkpath, with_metaclass
@@ -38,33 +39,62 @@ from barman.utils import fsync_dir, mkpath, with_metaclass
 _logger = logging.getLogger(__name__)
 
 
-class WalArchiverBatch(list):
-    def __init__(self, items, errors=None, skip=None):
+class WalArchiverQueue(list):
+    def __init__(self, items, errors=None, skip=None, batch_size=0):
         """
-        A WalArchiverBatch is a list of WalFileInfo which has two extra
+        A WalArchiverQueue is a list of WalFileInfo which has two extra
         attribute list:
 
         * errors: containing a list of unrecognized files
         * skip: containing a list of skipped files.
 
+        It also stores batch run size information in case
+        it is requested by configuration, in order to limit the
+        number of WAL files that are processed in a single
+        run of the archive-wal command.
+
         :param items: iterable from which initialize the list
+        :param batch_size: size of the current batch run (0=unlimited)
         :param errors: an optional list of unrecognized files
         :param skip: an optional list of skipped files
         """
-        super(WalArchiverBatch, self).__init__(items)
+        super(WalArchiverQueue, self).__init__(items)
         self.skip = []
         self.errors = []
         if skip is not None:
             self.skip = skip
         if errors is not None:
             self.errors = errors
+        # Normalises batch run size
+        if batch_size > 0:
+            self.batch_size = batch_size
+        else:
+            self.batch_size = 0
 
+    @property
+    def size(self):
+        """
+        Number of valid WAL segments waiting to be processed (in total)
 
-class ArchiverFailure(Exception):
-    """
-    Exception representing a failure during the execution
-    of the archive process
-    """
+        :return int: total number of valid WAL files
+        """
+        return len(self)
+
+    @property
+    def run_size(self):
+        """
+        Number of valid WAL files to be processed in this run - takes
+        in consideration the batch size
+
+        :return int: number of valid WAL files for this batch run
+        """
+        # In case a batch size has been explicitly specified
+        # (i.e. batch_size > 0), returns the minimum number between
+        # batch size and the queue size. Otherwise, simply
+        # returns the total queue size (unlimited batch size).
+        if self.batch_size > 0:
+            return min(self.size, self.batch_size)
+        return self.size
 
 
 class WalArchiver(with_metaclass(ABCMeta, RemoteStatusMixin)):
@@ -96,33 +126,70 @@ class WalArchiver(with_metaclass(ABCMeta, RemoteStatusMixin)):
         :raise ArchiverFailure: when something goes wrong
         """
 
-    def archive(self, fxlogdb, verbose=True):
+    def archive(self, verbose=True):
         """
         Archive WAL files, discarding duplicates or those that are not valid.
 
-        :param file fxlogdb: File object for xlogdb interactions
         :param boolean verbose: Flag for verbose output
         """
         compressor = self.backup_manager.compression_manager.get_compressor()
         stamp = datetime.datetime.utcnow().strftime('%Y%m%dT%H%M%SZ')
-        found = False
-        if verbose:
-            output.info("Processing xlog segments from %s for %s",
-                        self.name,
-                        self.config.name,
-                        log=False)
+        processed = 0
+        header = "Processing xlog segments from %s for %s" % (
+                 self.name, self.config.name)
+
+        # Get the next batch of WAL files to be processed
         batch = self.get_next_batch()
+
+        # Analyse the batch and properly log the information
+        if batch.size:
+            if batch.size > batch.run_size:
+                # Batch mode enabled
+                _logger.info("Found %s xlog segments from %s for %s."
+                             " Archive a batch of %s segments in this run.",
+                             batch.size,
+                             self.name,
+                             self.config.name,
+                             batch.run_size)
+                header += " (batch size: %s)" % batch.run_size
+            else:
+                # Single run mode (traditional)
+                _logger.info("Found %s xlog segments from %s for %s."
+                             " Archive all segments in one run.",
+                             batch.size,
+                             self.name,
+                             self.config.name)
+        else:
+            _logger.info("No xlog segments found from %s for %s.",
+                         self.name,
+                         self.config.name)
+
+        # Print the header (verbose mode)
+        if verbose:
+            output.info(header, log=False)
+
+        # Loop through all available WAL files
         for wal_info in batch:
-            if not found and not verbose:
-                output.info("Processing xlog segments from %s for %s",
-                            self.name,
-                            self.config.name,
-                            log=False)
-            found = True
+            # Print the header (non verbose mode)
+            if not processed and not verbose:
+                output.info(header, log=False)
+
+            # Exit when archive batch size is reached
+            if processed >= batch.run_size:
+                _logger.debug("Batch size reached (%s) - "
+                              "Exit %s process for %s",
+                              batch.batch_size,
+                              self.name,
+                              self.config.name)
+                break
+
+            processed += 1
 
             # Report to the user the WAL file we are archiving
             output.info("\t%s", wal_info.name, log=False)
-            _logger.info("Archiving %s/%s", self.config.name, wal_info.name)
+            _logger.info("Archiving segment %s of %s from %s: %s/%s",
+                         processed, batch.run_size, self.name,
+                         self.config.name, wal_info.name)
             # Archive the WAL file
             try:
                 self.archive_wal(compressor, wal_info)
@@ -150,26 +217,39 @@ class WalArchiver(with_metaclass(ABCMeta, RemoteStatusMixin)):
                                                 wal_info.name,
                                                 e))
                 return
-            # Updates the information of the WAL archive with
-            # the latest segments
-            fxlogdb.write(wal_info.to_xlogdb_line())
-            # flush and fsync for every line
-            fxlogdb.flush()
-            os.fsync(fxlogdb.fileno())
-        if not found and verbose:
+
+        if processed:
+            _logger.debug("Archived %s out of %s xlog segments from %s for %s",
+                          processed, batch.size, self.name, self.config.name)
+        elif verbose:
             output.info("\tno file found", log=False)
+
         if batch.errors:
             output.info("Some unknown objects have been found while "
                         "processing xlog segments for %s. "
                         "Objects moved to errors directory:",
                         self.config.name,
                         log=False)
+            # Log unexpected files
+            _logger.warning("Archiver is about to move %s unexpected file(s) "
+                            "to errors directory for %s from %s",
+                            len(batch.errors),
+                            self.config.name,
+                            self.name)
             for error in batch.errors:
-                output.info("\t%s", error)
+                basename = os.path.basename(error)
+                output.info("\t%s", basename, log=False)
+                # Print informative log line.
+                _logger.warning("Moving unexpected file for %s from %s: %s",
+                                self.config.name, self.name, basename)
                 error_dst = os.path.join(
                     self.config.errors_directory,
-                    "%s.%s.unknown" % (os.path.basename(error), stamp))
-                shutil.move(error, error_dst)
+                    "%s.%s.unknown" % (basename, stamp))
+                try:
+                    shutil.move(error, error_dst)
+                except IOError as e:
+                    if e.errno == errno.ENOENT:
+                        _logger.warning('%s not found' % error)
 
     def archive_wal(self, compressor, wal_info):
         """
@@ -235,36 +315,48 @@ class WalArchiver(with_metaclass(ABCMeta, RemoteStatusMixin)):
             # Compress the file only if not already compressed
             if compressor and not wal_info.compression:
                 compressor.compress(src_file, tmp_file)
-                shutil.copystat(src_file, tmp_file)
-                os.rename(tmp_file, dst_file)
-                os.unlink(src_file)
-                # Update wal_info
-                stat = os.stat(dst_file)
-                wal_info.size = stat.st_size
-                wal_info.compression = compressor.compression
-            else:
-                # Try to atomically rename the file. If successful,
-                # the renaming will be an atomic operation
-                # (this is a POSIX requirement).
-                try:
-                    os.rename(src_file, dst_file)
-                except OSError:
-                    # Source and destination are probably on different
-                    # filesystems
-                    shutil.copy2(src_file, tmp_file)
+
+            # Perform the real filesystem operation with the xlogdb lock taken.
+            # This makes the operation atomic from the xlogdb file POV
+            with self.server.xlogdb('a') as fxlogdb:
+                if compressor and not wal_info.compression:
+                    shutil.copystat(src_file, tmp_file)
                     os.rename(tmp_file, dst_file)
                     os.unlink(src_file)
-            # At this point the original file has been removed
-            wal_info.orig_filename = None
+                    # Update wal_info
+                    stat = os.stat(dst_file)
+                    wal_info.size = stat.st_size
+                    wal_info.compression = compressor.compression
+                else:
+                    # Try to atomically rename the file. If successful,
+                    # the renaming will be an atomic operation
+                    # (this is a POSIX requirement).
+                    try:
+                        os.rename(src_file, dst_file)
+                    except OSError:
+                        # Source and destination are probably on different
+                        # filesystems
+                        shutil.copy2(src_file, tmp_file)
+                        os.rename(tmp_file, dst_file)
+                        os.unlink(src_file)
+                # At this point the original file has been removed
+                wal_info.orig_filename = None
+
+                # Execute fsync() on the archived WAL file
+                file_fd = os.open(dst_file, os.O_RDONLY)
+                os.fsync(file_fd)
+                os.close(file_fd)
+                # Execute fsync() on the archived WAL containing directory
+                fsync_dir(dst_dir)
+                # Execute fsync() also on the incoming directory
+                fsync_dir(src_dir)
+                # Updates the information of the WAL archive with
+                # the latest segments
+                fxlogdb.write(wal_info.to_xlogdb_line())
+                # flush and fsync for every line
+                fxlogdb.flush()
+                os.fsync(fxlogdb.fileno())
 
-            # Execute fsync() on the archived WAL file
-            file_fd = os.open(dst_file, os.O_RDONLY)
-            os.fsync(file_fd)
-            os.close(file_fd)
-            # Execute fsync() on the archived WAL containing directory
-            fsync_dir(dst_dir)
-            # Execute fsync() also on the incoming directory
-            fsync_dir(src_dir)
         except Exception as e:
             # In case of failure save the exception for the post scripts
             error = e
@@ -295,9 +387,9 @@ class WalArchiver(with_metaclass(ABCMeta, RemoteStatusMixin)):
     @abstractmethod
     def get_next_batch(self):
         """
-        Return a WalArchiverBatch containing the WAL files to be archived.
+        Return a WalArchiverQueue containing the WAL files to be archived.
 
-        :rtype: WalArchiverBatch
+        :rtype: WalArchiverQueue
         """
 
     @abstractmethod
@@ -310,6 +402,12 @@ class WalArchiver(with_metaclass(ABCMeta, RemoteStatusMixin)):
              of the results of the various checks
         """
 
+    @abstractmethod
+    def status(self):
+        """
+        Set additional status info - invoked by Server.status()
+        """
+
     @staticmethod
     def summarise_error_files(error_files):
         """
@@ -360,7 +458,7 @@ class FileWalArchiver(WalArchiver):
         """
         result = dict.fromkeys(
             ['archive_mode', 'archive_command'], None)
-        postgres = self.backup_manager.server.postgres
+        postgres = self.server.postgres
         # If Postgres is not available we cannot detect anything
         if not postgres:
             return result
@@ -372,7 +470,6 @@ class FileWalArchiver(WalArchiver):
         pg_stat_archiver = postgres.get_archiver_stats()
         if pg_stat_archiver is not None:
             result.update(pg_stat_archiver)
-
         return result
 
     def get_next_batch(self):
@@ -380,8 +477,10 @@ class FileWalArchiver(WalArchiver):
         Returns the next batch of WAL files that have been archived through
         a PostgreSQL's 'archive_command' (in the 'incoming' directory)
 
-        :return: WalArchiverBatch: list of WAL files
+        :return: WalArchiverQueue: list of WAL files
         """
+        # Get the batch size from configuration (0 = unlimited)
+        batch_size = self.config.archiver_batch_size
         # List and sort all files in the incoming directory
         file_names = glob(os.path.join(
             self.config.incoming_wals_directory, '*'))
@@ -399,7 +498,9 @@ class FileWalArchiver(WalArchiver):
 
         # Build the list of WalFileInfo
         wal_files = [WalFileInfo.from_file(f) for f in files]
-        return WalArchiverBatch(wal_files, errors=errors)
+        return WalArchiverQueue(wal_files,
+                                batch_size=batch_size,
+                                errors=errors)
 
     def check(self, check_strategy):
         """
@@ -409,35 +510,79 @@ class FileWalArchiver(WalArchiver):
         :param CheckStrategy check_strategy: the strategy for the management
              of the results of the various checks
         """
+        check_strategy.init_check('archive_mode')
         remote_status = self.get_remote_status()
         # If archive_mode is None, there are issues connecting to PostgreSQL
         if remote_status['archive_mode'] is None:
             return
         # Check archive_mode parameter: must be on
         if remote_status['archive_mode'] in ('on', 'always'):
-            check_strategy.result(self.config.name, 'archive_mode', True)
+            check_strategy.result(self.config.name, True)
         else:
             msg = "please set it to 'on'"
             if self.server.postgres.server_version >= 90500:
                 msg += " or 'always'"
-            check_strategy.result(self.config.name, 'archive_mode', False, msg)
-
+            check_strategy.result(self.config.name, False, hint=msg)
+        check_strategy.init_check('archive_command')
         if remote_status['archive_command'] and \
                 remote_status['archive_command'] != '(disabled)':
-            check_strategy.result(self.config.name, 'archive_command',
-                                  True)
+            check_strategy.result(self.config.name,
+                                  True,
+                                  check='archive_command')
 
             # Report if the archiving process works without issues.
             # Skip if the archive_command check fails
             # It can be None if PostgreSQL is older than 9.4
             if remote_status.get('is_archiving') is not None:
                 check_strategy.result(
-                    self.config.name, 'continuous archiving',
-                    remote_status['is_archiving'])
+                    self.config.name,
+                    remote_status['is_archiving'],
+                    check='continuous archiving')
         else:
             check_strategy.result(
-                self.config.name, 'archive_command', False,
-                'please set it accordingly to documentation')
+                self.config.name, False,
+                hint='please set it accordingly to documentation')
+
+    def status(self):
+        """
+        Set additional status info - invoked by Server.status()
+        """
+        remote_status = self.get_remote_status()
+
+        # If archive_mode is None, there are issues connecting to PostgreSQL
+        if remote_status['archive_mode'] is None:
+            return
+
+        output.result('status', self.config.name,
+                      "archive_command",
+                      "PostgreSQL 'archive_command' setting",
+                      remote_status['archive_command'] or
+                      "FAILED (please set it accordingly to documentation)")
+        last_wal = remote_status.get('last_archived_wal')
+        # If PostgreSQL is >= 9.4 we have the last_archived_time
+        if last_wal and remote_status.get('last_archived_time'):
+                last_wal += ", at %s" % (
+                    remote_status['last_archived_time'].ctime())
+        output.result('status', self.config.name,
+                      "last_archived_wal",
+                      "Last archived WAL",
+                      last_wal or "No WAL segment shipped yet")
+        # Set output for WAL archive failures (PostgreSQL >= 9.4)
+        if remote_status.get('failed_count') is not None:
+            remote_fail = str(remote_status['failed_count'])
+            if int(remote_status['failed_count']) > 0:
+                remote_fail += " (%s at %s)" % (
+                    remote_status['last_failed_wal'],
+                    remote_status['last_failed_time'].ctime())
+            output.result('status', self.config.name, 'failed_count',
+                          'Failures of WAL archiver', remote_fail)
+        # Add hourly archive rate if available (PostgreSQL >= 9.4) and > 0
+        if remote_status.get('current_archived_wals_per_second'):
+            output.result(
+                'status', self.config.name,
+                'server_archived_wals_per_hour',
+                'Server WAL archiving rate', '%0.2f/hour' % (
+                    3600 * remote_status['current_archived_wals_per_second']))
 
 
 class StreamingWalArchiver(WalArchiver):
@@ -457,71 +602,70 @@ class StreamingWalArchiver(WalArchiver):
 
         :rtype: dict[str, None|str]
         """
-        result = dict.fromkeys(
+        remote_status = dict.fromkeys(
             ('pg_receivexlog_compatible',
              'pg_receivexlog_installed',
              'pg_receivexlog_path',
+             'pg_receivexlog_supports_slots',
+             'pg_receivexlog_synchronous',
              'pg_receivexlog_version'),
             None)
 
-        # Check the server version from the streaming
-        # connection
-        streaming = self.backup_manager.server.streaming
-        server_txt_version = None
-        if streaming:
-            server_txt_version = streaming.server_txt_version
-        if server_txt_version:
-            pg_version = Version(utils.simplify_version(server_txt_version))
-        else:
-            # No log here, it has already been logged in the
-            # StreamingConnection class or during the Server initialization
-            pg_version = None
-
-        # Detect a pg_receivexlog executable
-        pg_receivexlog = utils.which("pg_receivexlog",
-                                     self.backup_manager.server.path)
-
         # Test pg_receivexlog existence
-        if pg_receivexlog:
-            result["pg_receivexlog_installed"] = True
-            result["pg_receivexlog_path"] = pg_receivexlog
+        version_info = PgReceiveXlog.get_version_info(
+            self.server.path)
+        if version_info['full_path']:
+            remote_status["pg_receivexlog_installed"] = True
+            remote_status["pg_receivexlog_path"] = version_info['full_path']
+            remote_status["pg_receivexlog_version"] = (
+                version_info['full_version'])
+            pgreceivexlog_version = version_info['major_version']
         else:
-            result["pg_receivexlog_installed"] = False
-            return result
+            remote_status["pg_receivexlog_installed"] = False
+            return remote_status
 
-        receivexlog = Command(pg_receivexlog, check=True)
-
-        # Obtain the `pg_receivexlog` version
-        try:
-            receivexlog("--version")
-            splitter_version = receivexlog.out.strip().split()
-            result["pg_receivexlog_version"] = splitter_version[-1]
-            receivexlog_version = Version(
-                utils.simplify_version(result["pg_receivexlog_version"]))
-        except CommandFailedException as e:
-            receivexlog_version = None
-            _logger.debug("Error invoking pg_receivexlog: %s", e)
+        # Retrieve the PostgreSQL version
+        pg_version = None
+        if self.server.streaming is not None:
+            pg_version = self.server.streaming.server_major_version
 
         # If one of the version is unknown we cannot compare them
-        if receivexlog_version is None or pg_version is None:
-            return result
+        if pgreceivexlog_version is None or pg_version is None:
+            return remote_status
+
+        # pg_version is not None so transform into a Version object
+        # for easier comparison between versions
+        pg_version = Version(pg_version)
+
+        # Set conservative default values (False) for modern features
+        remote_status["pg_receivexlog_compatible"] = False
+        remote_status['pg_receivexlog_supports_slots'] = False
+        remote_status["pg_receivexlog_synchronous"] = False
 
         # pg_receivexlog 9.2 is compatible only with PostgreSQL 9.2.
-        if "9.2" == pg_version == receivexlog_version:
-            result["pg_receivexlog_compatible"] = True
+        if "9.2" == pg_version == pgreceivexlog_version:
+            remote_status["pg_receivexlog_compatible"] = True
 
         # other versions are compatible with lesser versions of PostgreSQL
         # WARNING: The development versions of `pg_receivexlog` are considered
         # higher than the stable versions here, but this is not an issue
         # because it accepts everything that is less than
         # the `pg_receivexlog` version(e.g. '9.6' is less than '9.6devel')
-        elif "9.2" < pg_version <= receivexlog_version:
-            result["pg_receivexlog_compatible"] = True
+        elif "9.2" < pg_version <= pgreceivexlog_version:
+            # At least PostgreSQL 9.3 is required here
+            remote_status["pg_receivexlog_compatible"] = True
 
-        else:
-            result["pg_receivexlog_compatible"] = False
+            # replication slots are supported starting from version 9.4
+            if "9.4" <= pg_version <= pgreceivexlog_version:
+                remote_status['pg_receivexlog_supports_slots'] = True
 
-        return result
+            # Synchronous WAL streaming requires replication slots
+            # and pg_receivexlog >= 9.5
+            if "9.4" <= pg_version and "9.5" <= pgreceivexlog_version:
+                remote_status["pg_receivexlog_synchronous"] = (
+                    self._is_synchronous())
+
+        return remote_status
 
     def receive_wal(self, reset=False):
         """
@@ -540,11 +684,12 @@ class StreamingWalArchiver(WalArchiver):
             return
 
         # Execute basic sanity checks on PostgreSQL connection
-        postgres_status = self.server.streaming.get_remote_status()
-        if postgres_status["streaming_supported"] is None:
+        streaming_status = self.server.streaming.get_remote_status()
+        if streaming_status["streaming_supported"] is None:
             raise ArchiverFailure(
-                'failed opening the PostgreSQL streaming connection')
-        elif not postgres_status["streaming_supported"]:
+                'failed opening the PostgreSQL streaming connection '
+                'for server %s' % (self.config.name))
+        elif not streaming_status["streaming_supported"]:
             raise ArchiverFailure(
                 'PostgreSQL version too old (%s < 9.2)' %
                 self.server.streaming.server_txt_version)
@@ -558,24 +703,63 @@ class StreamingWalArchiver(WalArchiver):
                 'pg_receivexlog version not compatible with '
                 'PostgreSQL server version')
 
+        # Execute sanity check on replication slot usage
+        if self.config.slot_name:
+            # Check if slots are supported
+            if not remote_status['pg_receivexlog_supports_slots']:
+                raise ArchiverFailure(
+                    'Physical replication slot not supported by %s '
+                    '(9.4 or higher is required)' %
+                    self.server.streaming.server_txt_version)
+            # Check if the required slot exists
+            postgres_status = self.server.postgres.get_remote_status()
+            if postgres_status['replication_slot'] is None:
+                raise ArchiverFailure(
+                    "replication slot '%s' doesn't exist. "
+                    "Please execute "
+                    "'barman receive-wal --create-slot %s'" %
+                    (self.config.slot_name, self.config.name))
+            # Check if the required slot is available
+            if postgres_status['replication_slot'].active:
+                raise ArchiverFailure(
+                    "replication slot '%s' is already in use" %
+                    (self.config.slot_name,))
+
         # Make sure we are not wasting precious PostgreSQL resources
-        self.server.postgres.close()
-        self.server.streaming.close()
+        self.server.close()
 
         _logger.info('Activating WAL archiving through streaming protocol')
         try:
             output_handler = PgReceiveXlog.make_output_handler(
                 self.config.name + ': ')
-            receive = PgReceiveXlog(remote_status['pg_receivexlog_path'],
-                                    self.config.streaming_conninfo,
-                                    self.config.streaming_wals_directory,
-                                    out_handler=output_handler,
-                                    err_handler=output_handler)
+            receive = PgReceiveXlog(
+                connection=self.server.streaming,
+                destination=self.config.streaming_wals_directory,
+                command=remote_status['pg_receivexlog_path'],
+                version=remote_status['pg_receivexlog_version'],
+                app_name=self.config.streaming_archiver_name,
+                path=self.server.path,
+                slot_name=self.config.slot_name,
+                synchronous=remote_status['pg_receivexlog_synchronous'],
+                out_handler=output_handler,
+                err_handler=output_handler
+            )
+            # Finally execute the pg_receivexlog process
             receive.execute()
         except CommandFailedException as e:
-            _logger.error(e)
-            raise ArchiverFailure("pg_receivexlog exited with an error. "
-                                  "Check the logs for more information.")
+            # Retrieve the return code from the exception
+            ret_code = e.args[0]['ret']
+            if ret_code < 0:
+                # If the return code is negative, then pg_receivexlog
+                # was terminated by a signal
+                msg = "pg_receivexlog terminated by signal: %s" \
+                      % abs(ret_code)
+            else:
+                # Otherwise terminated with an error
+                msg = "pg_receivexlog terminated with error code: %s"\
+                      % ret_code
+
+            raise ArchiverFailure(msg)
         except KeyboardInterrupt:
             # This is a normal termination, so there is nothing to do beside
             # informing the user.
@@ -601,30 +785,38 @@ class StreamingWalArchiver(WalArchiver):
         because the 'pg_receivexlog' process needs at least one file to
         detect the current streaming position after a restart.
 
-        :return: WalArchiverBatch: list of WAL files
+        :return: WalArchiverQueue: list of WAL files
         """
+        # Get the batch size from configuration (0 = unlimited)
+        batch_size = self.config.streaming_archiver_batch_size
         # List and sort all files in the incoming directory
         file_names = glob(os.path.join(
             self.config.streaming_wals_directory, '*'))
         file_names.sort()
 
         # Process anything that looks like a valid WAL file,
-        # including partial ones.
+        # including partial ones and history files.
         # Anything else is treated like an error/anomaly
         files = []
         skip = []
         errors = []
         for file_name in file_names:
-            if xlog.is_wal_file(file_name) and os.path.isfile(file_name):
-                files.append(file_name)
-            elif xlog.is_partial_file(file_name) and os.path.isfile(file_name):
+            if not os.path.isfile(file_name):
+                errors.append(file_name)
+            elif xlog.is_partial_file(file_name):
                 skip.append(file_name)
+            elif xlog.is_any_xlog_file(file_name):
+                files.append(file_name)
             else:
                 errors.append(file_name)
         # In case of more than a partial file, keep the last
-        # and treat the rest as errors
+        # and treat the rest as normal files
         if len(skip) > 1:
-            errors.extend(skip[:-1])
+            partials = skip[:-1]
+            _logger.info('Archiving partial files for server %s: %s' %
+                         (self.config.name,
+                          ", ".join([os.path.basename(f) for f in partials])))
+            files.extend(partials)
             skip = skip[-1:]
 
         # Keep the last full WAL file in case no partial file is present
@@ -633,31 +825,38 @@ class StreamingWalArchiver(WalArchiver):
 
         # Build the list of WalFileInfo
         wal_files = [WalFileInfo.from_file(f, compression=None) for f in files]
-        return WalArchiverBatch(wal_files, errors=errors, skip=skip)
+        return WalArchiverQueue(wal_files,
+                                batch_size=batch_size,
+                                errors=errors, skip=skip)
 
     def check(self, check_strategy):
         """
-        Perform additional checks for FileWalArchiver - invoked
+        Perform additional checks for StreamingWalArchiver - invoked
         by server.check_postgres
 
         :param CheckStrategy check_strategy: the strategy for the management
              of the results of the various checks
         """
 
+        check_strategy.init_check('pg_receivexlog')
         # Check the version of pg_receivexlog
         remote_status = self.get_remote_status()
         check_strategy.result(
-            self.config.name, 'pg_receivexlog',
+            self.config.name,
             remote_status['pg_receivexlog_installed'])
         hint = None
+        check_strategy.init_check('pg_receivexlog compatible')
         if not remote_status['pg_receivexlog_compatible']:
+            pg_version = 'Unknown'
+            if self.server.streaming is not None:
+                pg_version = self.server.streaming.server_txt_version
             hint = "PostgreSQL version: %s, pg_receivexlog version: %s" % (
-                self.server.streaming.server_txt_version,
+                pg_version,
                 remote_status['pg_receivexlog_version']
             )
-        check_strategy.result(
-            self.config.name, 'pg_receivexlog compatible',
-            remote_status['pg_receivexlog_compatible'], hint=hint)
+        check_strategy.result(self.config.name,
+                              remote_status['pg_receivexlog_compatible'],
+                              hint=hint)
 
         # Check if pg_receivexlog is running, by retrieving a list
         # of running 'receive-wal' processes from the process manager.
@@ -665,10 +864,50 @@ class StreamingWalArchiver(WalArchiver):
 
         # If there's at least one 'receive-wal' process running for this
         # server, the test is passed
+        check_strategy.init_check('receive-wal running')
         if receiver_list:
             check_strategy.result(
-                self.config.name, 'receive-wal running', True)
+                self.config.name, True)
         else:
             check_strategy.result(
-                self.config.name, 'receive-wal running', False,
-                'See the Barman log file for more details')
+                self.config.name,
+                False,
+                hint='See the Barman log file for more details')
+
+    def _is_synchronous(self):
+        """
+        Check if receive-wal process is eligible for synchronous replication
+
+        The receive-wal process is eligible for synchronous replication
+        if `synchronous_standby_names` is configured and contains
+        the value of `streaming_archiver_name`
+
+        :rtype: bool
+        """
+        # Nothing to do if postgres connection is not working
+        postgres = self.server.postgres
+        if postgres is None or postgres.server_txt_version is None:
+            return None
+
+        # Check if synchronous WAL streaming can be enabled
+        # by peeking 'synchronous_standby_names'
+        postgres_status = postgres.get_remote_status()
+        syncnames = postgres_status['synchronous_standby_names']
+        _logger.debug("Look for '%s' in "
+                      "'synchronous_standby_names': %s",
+                      self.config.streaming_archiver_name, syncnames)
+        # The receive-wal process is eligible for synchronous replication
+        # if `synchronous_standby_names` is configured and contains
+        # the value of `streaming_archiver_name`
+        synchronous = (syncnames and
+                       self.config.streaming_archiver_name in syncnames)
+        _logger.info('Synchronous WAL streaming for %s: %s',
+                     self.config.streaming_archiver_name,
+                     synchronous)
+        return synchronous
+
+    def status(self):
+        """
+        Set additional status info - invoked by Server.status()
+        """
+        # TODO: Add status information for WAL streaming
